import json
from django.views import View
from django.db.models import Q

from apps.ebpf.models import EbpfEvent
from libs import json_response, auth
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger

ITEMS_PER_PAGE = 10

class EbpfEventView(View):

    def get(self, request):
        # 获取查询参数
        ip = request.GET.get('ip')
        start_time = request.GET.get('start_time')
        end_time = request.GET.get('end_time')
        event_type = request.GET.get('event_type')
        event = request.GET.get('event')
        severity_level = request.GET.get('severity_level')
        page = int(request.GET.get('page', 1))  # 获取当前页数，默认为第1页
        size = int(request.GET.get('size', ITEMS_PER_PAGE))

        # 构建查询条件
        conditions = []
        if ip:
            conditions.append(Q(ip__contains=ip))
        if start_time and end_time:
            conditions.append(Q(updated_at__gte=start_time))
            conditions.append(Q(updated_at__lte=end_time))
        if start_time and not end_time:
            conditions.append(Q(updated_at__gte=start_time))
        if not start_time and end_time:
            conditions.append(Q(updated_at__lte=end_time))
        if event_type:
            conditions.append(Q(event_type__contains=event_type))
        if severity_level:
            conditions.append(Q(severity_level=severity_level))
        if event:
            conditions.append(Q(event__contains=event))

        # 执行查询，按time倒序，ip正序排序
        event_list = EbpfEvent.objects.filter(*conditions).order_by('-updated_at', 'ip')

        # 创建一个 Paginator 对象，提供预先计算好的计数值
        paginator = Paginator(event_list, size) 
        total = paginator.count

        # 获取该页的对象，执行数据库查询
        events = paginator.get_page(page)

        # 序列化查询结果
        events = ([{"id": event.id, 
                    "ip": event.ip, 
                    "user": event.user, 
                    "created_at": event.created_at, 
                    "updated_at": event.updated_at, 
                    "severity_level": event.severity_level, 
                    "event_type": event.event_type, 
                    "event": event.event} 
                   for event in events])

        # 计算总页数（假设数据库中总条数已知）
        total_pages = (total + size - 1) // size

        # 返回结果，包括分页信息
        response_data = {
            "events": events,
            "total": total,
            "total_pages": total_pages,
            "current_page": page,
        }

        return json_response(response_data)

    def post(self, request):
        if request.method == 'POST':
            json_data = json.loads(request.body)
            try:
                EbpfEvent.objects.create(
                            ip=json_data['ip'],
                            user=json_data['user'],
                            created_at=json_data['created_at'],
                            updated_at=json_data['updated_at'],
                            severity_level=json_data['severity_level'],
                            event_hash=json_data['event_hash'],
                            event_type=json_data['event_type'],
                            event=json_data['event'])
                return json_response()
            except Exception as e:
                return json_response(error=e)
        else:
            return json_response(error='method not allowed')

def update_by_id(request, id):
    if request.method == 'PATCH':
        json_data = json.loads(request.body)
        try:
            EbpfEvent.objects.filter(id=id).update(
                        ip=json_data['ip'],
                        user=json_data['user'],
                        # created_at=json_data['created_at'],
                        updated_at=json_data['updated_at'],
                        severity_level=json_data['severity_level'],
                        event_hash=json_data['event_hash'],
                        event_type=json_data['event_type'],
                        event=json_data['event'])
            return json_response()
        except Exception as e:
            return json_response(error=e)
    else:
        return json_response(error='method not allowed')

def find_by_ip_and_hash(request):
    if request.method == 'GET':
        # 获取查询参数
        ip = request.GET.get('ip')
        event_hash = request.GET.get('hash')
        if not ip or not event_hash:
            return json_response(error='ip or event_hash is required')

        try:
            ebpf_event = EbpfEvent.objects.filter(ip=ip, event_hash=event_hash).first()
            if not ebpf_event:
                return json_response(error='event not found')
            return json_response(ebpf_event.to_dict())
        except Exception as e:
            return json_response(error=e)
    else:
        return json_response(error='method not allowed')
