from concurrent import futures
import json
import os
import re
from typing import List
from apps.patch.models import Cve, HostScan, HostScanCve, HostScanPatch, Patch
from apps.host.models import Host, HostExtend
from apps.host.utils import get_ssh
from apps.setting.utils import AppSetting
from django_redis import get_redis_connection
from libs import human_datetime
import logging

logger = logging.getLogger('patches')

def import_patch(patches_data):
    for patch_info in patches_data:
        cve = patch_info['cve_list'][0]
        sa_cvss = patch_info.get('sa_cvss', 0)
        sa_id = patch_info.get('sa_id', '')
        sa_link = patch_info.get('sa_link', '')
        sa_title = patch_info.get('sa_title', '')
        sa_sammary = patch_info.get('sa_summary', '')
        sa_release_time = patch_info.get('sa_release_time', '')

        # 更新数据
        updated_count = Cve.objects.filter(cve=cve).update(
            sa_id=sa_id,
            sa_link=sa_link,
            sa_title=sa_title,
            sa_summary=sa_sammary,
            sa_cvss=sa_cvss,
            sa_release_time=sa_release_time,
        )

        # 如果没有历史数据，则创建
        if updated_count == 0:
            Cve.objects.create(
                cve=cve,
                sa_id=sa_id,
                sa_link=sa_link,
                sa_title=sa_title,
                sa_summary=sa_sammary,
                sa_cvss=sa_cvss,
                sa_release_time=sa_release_time
            )
        
        # 收集漏洞补丁信息
        patches = []
        # 删除历史数据
        Patch.objects.filter(cve=cve).delete()
        for platform, packages in patch_info['packages'].items():
            for pkg in packages:
                patches.append(Patch(
                    cve=cve,
                    platform=platform,
                    arch=pkg['arch'],
                    package_name=pkg['name'],
                    package_file_name=pkg['rpm'],
                    version=pkg['version'],
                    cvss=sa_cvss,
                    updated_at=human_datetime(),  
                    created_at=human_datetime(),
                ))
        
        # 批量创建漏洞补丁信息
        Patch.objects.bulk_create(patches)

def do_batch_host_scan(token, hosts: List[Host]): 
    private_key, public_key = AppSetting.get_ssh_key()
    threads, rds = [], get_redis_connection()
    max_workers = max(10, os.cpu_count() * 5)
    with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        for host in hosts:
            host_extend = HostExtend.objects.filter(host_id=host.id).first()
            t = executor.submit(_batch_host_scan, host, host_extend, private_key, public_key)
            t.host = host
            threads.append(t)
        for t in futures.as_completed(threads):
            exception = t.exception()
            if exception:
                logging.error(f'{exception}')
                rds.rpush(token, json.dumps({'key': t.host.id, 'status': 'fail', 'message': f'{exception}'}))
            else:
                rds.rpush(token, json.dumps({'key': t.host.id, 'status': 'ok'}))
        rds.expire(token, 600)

def _batch_host_scan(host, host_extend, private_key, public_key): 
    platform = host_extend.platform
    arch = host_extend.arch

    try:
        # 删除未处理的历史数据
        host_scans = HostScan.objects.filter(host_id=host.id, is_fixed=False)
        for host_scan in host_scans:
            host_scan.delete()

        # 创建主机扫描记录
        host_scan = (HostScan.objects.create(
            host_id=host.id, 
            created_at=human_datetime(), 
            updated_at=human_datetime(), 
            is_fixed=False))
    except Exception:
        raise
    
    private_key, public_key = AppSetting.get_ssh_key()
    kwargs = host.to_dict(selects=('hostname', 'port', 'username'))
    ssh = get_ssh(kwargs, pkey=host.pkey, private_key=private_key, public_key=public_key, password=None)

    code, out = ssh.exec_command_raw("rpm -qa --queryformat '%{NAME}@%{VERSION}-%{RELEASE}\n'")
    if code != 0:
        raise Exception(out)
    else:
        packages = out.split('\n') 

    # 分离名称和版本信息
    packages = [package.split('@') for package in packages]
    packages = [{'name': package[0], 'version': package[1]} for package in packages]

    # 按照平台和架构获取补丁信息
    patches = (
        Patch.objects
        .filter(platform=platform, arch=arch)
        .values('cve', 'package_name', 'version', 'cvss'))
    
    # 按照 package_name 分组
    grouped_patches_dict = {}
    for patch in patches:
        if patch['package_name'] not in grouped_patches_dict:
            grouped_patches_dict[patch['package_name']] = []
        grouped_patches_dict[patch['package_name']].append(patch)

    host_scan_patch_list = []
    # 遍历 packages，对比 version 检查是否需要更新
    for package in packages:
        host_scan_patch = HostScanPatch()
        host_scan_patch.package_name = package['name']
        host_scan_patch.current_version = package['version']
        host_scan_patch.target_version = package['version']
        host_scan_patch.is_fixed = False
        host_scan_patch.created_at = human_datetime()
        host_scan_patch.updated_at = human_datetime()
        host_scan_patch.cves = []
        if package['name'] in grouped_patches_dict:
            # 跳过 openssh 相关的包
            if package['name'] in ['openssh', 'openssh-clients', 'openssh-server']:
                continue

            for patch in grouped_patches_dict[package['name']]:
                (logging.debug('current_version: %s target_version: %s, compare_version: %s' %
                       (host_scan_patch.current_version, 
                       patch['version'], 
                       compare_version(host_scan_patch.current_version, patch['version']))))
                if compare_version(host_scan_patch.current_version, patch['version']) < 0:
                    host_scan_patch.target_version = patch['version']
                    host_scan_patch.cvss = patch['cvss']
                    host_scan_patch.cve = patch['cve']
        if host_scan_patch.target_version != host_scan_patch.current_version:
            host_scan_patch_list.append(host_scan_patch)
    
    for host_scan_patch in host_scan_patch_list:
        host_scan_patch.hs_id = host_scan.id

    all_cves = set()
    for host_scan_patch in host_scan_patch_list:
        all_cves.add(host_scan_patch.cve)

    HostScanPatch.objects.bulk_create(host_scan_patch_list)

    cves = Cve.objects.all();
    cves_dict = {cve.cve: cve for cve in cves}

    host_scan_cve_list = ([HostScanCve(host_id=host.id, 
                                       hs_id=host_scan.id,
                                       cve=cve, 
                                       is_fixed=False, 
                                       created_at=human_datetime(), 
                                       updated_at=human_datetime(),
                                       cvss=cves_dict.get(cve).sa_cvss if cves_dict.get(cve) else None) 
                                       for cve in all_cves])
    HostScanCve.objects.bulk_create(host_scan_cve_list)


def compare_version(version1, version2):
    # Handle version format as epoch:version
    if ":" not in version1:
        version1 = "0:" + version1

    if ":" not in version2:
        version2 = "0:" + version2

    version1_parts = re.split('[:-]', version1)
    version2_parts = re.split('[:-]', version2)

    # Compare major version numbers
    result = compare_version_parts(version1_parts[1], version2_parts[1])

    # If major version numbers are the same, compare minor version numbers
    if result == 0:
        result = compare_version_parts(version1_parts[2], version2_parts[2])

    return result

def compare_version_parts(part1, part2):
    part1_digits = list(filter(None, re.split('\\D+', part1)))
    part2_digits = list(filter(None, re.split('\\D+', part2)))

    for i in range(min(len(part1_digits), len(part2_digits))):
        num1 = int(part1_digits[i])
        num2 = int(part2_digits[i])

        if num1 != num2:
            return (num1 > num2) - (num1 < num2)

    return (len(part1_digits) > len(part2_digits)) - (len(part1_digits) < len(part2_digits))

def do_batch_host_patch(token, hosts: List[Host], cves: List[str]): 

    private_key, public_key = AppSetting.get_ssh_key()
    threads, rds = [], get_redis_connection()
    max_workers = max(10, os.cpu_count() * 5)
    with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        for host in hosts:
            host_scan = HostScan.objects.filter(host_id=host.id, is_fixed=False).first()
            if not host_scan:
                continue
            if not cves:
                host_scan_patches = HostScanPatch.objects.filter(hs__id=host_scan.id)
            else:
                host_scan_patches = HostScanPatch.objects.filter(hs__id=host_scan.id, cve__in=cves)

            t = executor.submit(_host_patch, host, host_scan_patches, cves, private_key, public_key)
            t.host = host
            t.host_scan = host_scan
            t.host_scan_patches = host_scan_patches
            t.cves = cves
            threads.append(t)
        for t in futures.as_completed(threads):
            exception = t.exception()
            t.host_scan.updated_at = human_datetime()
            if exception:
                logger.error(f'{exception}')
                rds.rpush(token, json.dumps({'key': t.host.id, 'status': 'fail', 'message': f'{exception}'}))
                t.host_scan.patch_messages = f'{exception}'
            else:
                rds.rpush(token, json.dumps({'key': t.host.id, 'status': 'ok'}))
                if t.cves:
                    HostScanCve.objects.filter(hs_id=t.host_scan.id, cve__in=t.cves).update(is_fixed=True, updated_at=human_datetime())
                if t.host_scan_patches:
                    t.host_scan_patches.update(is_fixed=True, updated_at=human_datetime())

            count_not_fixed = HostScanPatch.objects.filter(hs__id=t.host_scan.id, is_fixed=False).count();
            if count_not_fixed == 0:
                t.host_scan.is_fixed = True
                t.host_scan.save()
        rds.expire(token, 600)
        

def _host_patch(host, host_scan_patches, cves, private_key=None, public_key=None):
    packages = [f'{patch.package_name}-{patch.target_version}' for patch in host_scan_patches]
    if not packages:
        return
    command = f'sudo yum install -y %s' % ' '.join(packages)
    logger.info('ssh host: %s, command: %s', host.hostname, command)
    kwargs = host.to_dict(selects=('hostname', 'port', 'username'))
    ssh = get_ssh(kwargs, host.pkey, private_key, public_key, None)
    code, out = ssh.exec_command_raw(command)
    if code != 0:
        raise Exception(f'{out.strip()}')


security_level_dict = {
    '1': (0.1, 3.9),
    '2': (4, 6.9),
    '3': (7, 8.9),
    '4': (9, 10),
}

def get_security_level(cvss):
    if cvss is None:
        return '0'
    if cvss < 4:
        return '1'
    elif cvss < 7:
        return '2'
    elif cvss < 9:
        return '3'
    else:
        return '4'