from threading import Thread
import uuid
from django.db.models import Count, Q
from django.views import View

from apps.host.models import Host
from apps.patch.models import Cve, HostScan, HostScanCve, HostScanPatch, Patch
from apps.account.utils import get_host_perms

from .utils import do_batch_host_patch, do_batch_host_scan, import_patch, security_level_dict, get_security_level
from libs import json_response, auth
import tarfile
import json

class HostView(View):
    def get(self, request):

        host_name = request.GET.get('host_name') or ''
        filter = Q(name__contains=host_name)

        if not request.user.is_supper:
            filter &= Q(id__in=get_host_perms(request.user))

        hosts = Host.objects.filter(filter).prefetch_related('groups')
        host_ids = [h.id for h in hosts]

        host_scans = (HostScan.objects
                          .filter(host_id__in=host_ids, is_fixed=False)
                          .values('host_id', 'created_at'))
        host_scans_dict = {x['host_id']: x['created_at'] for x in host_scans}

        host_cves_count = (HostScanCve.objects
                           .filter(is_fixed=False, hs__is_fixed=False)
                           .order_by('host_id')
                           .values('host_id')
                           .annotate(num_cves=Count('cve')))
        host_cves_count_dict = {x['host_id']: x['num_cves'] for x in host_cves_count}

        result = []

        for host in hosts:
            host_dict = host.to_dict();
            host_dict['group_names'] = ', '.join([g.name for g in host.groups.all()])
            host_dict['last_scan_at'] = host_scans_dict.get(host.id)
            host_dict['cve_count'] = host_cves_count_dict.get(host.id, 0)
            result.append(host_dict)
        
        return json_response(result)
        

class CveView(View):
    def get(self, request):
        security_level = request.GET.get('security_level')
        cve = request.GET.get('cve')
        if (security_level and security_level in security_level_dict):
            min_cvss, max_cvss = security_level_dict.get(security_level)
            cve_host_count = (HostScanCve.objects
                            .filter(cve__icontains=cve, 
                                    cvss__gte=min_cvss, 
                                    cvss__lte=max_cvss,
                                    hs__is_fixed=False,
                                    is_fixed=False)
                            .annotate(num_hosts=Count('host_id')).values('cve', 'num_hosts'))
            cves = (Cve.objects
                    .filter(cve__icontains=cve, 
                            sa_cvss__gte=min_cvss, 
                            sa_cvss__lte=max_cvss)
                    .values('cve', 'sa_release_time', 'sa_cvss'))
        else:
            cve_host_count = (HostScanCve.objects
                            .filter(cve__icontains=cve,
                                    hs__is_fixed=False,
                                    is_fixed=False)
                            .annotate(num_hosts=Count('host_id')).values('cve', 'num_hosts'))
            cves = Cve.objects.filter(cve__icontains=cve).values('cve', 'sa_release_time', 'sa_cvss')

        cve_host_count_dict = {x['cve']: x['num_hosts'] for x in cve_host_count}
        for cve in cves:
            cve['security_level'] = get_security_level(cve['sa_cvss'])
            cve['host_count'] = cve_host_count_dict.get(cve['cve'], 0)
        
        return json_response(list(cves))

    @auth('cve.cve.upload')
    def post(self, request):
        try:
            if request.method == 'POST' and request.FILES.get('file'):
                tarball = request.FILES['file']
                
                # 解压并读取tar.gz文件中的JSON文件
                patches_data = []
                with tarfile.open(fileobj=tarball, mode="r:gz") as tar:
                    for member in tar:
                        if member.isfile() and member.name.endswith('.json'):
                            with tar.extractfile(member) as file_obj:
                                json_bytes = file_obj.read()
                                json_str = json_bytes.decode('utf-8')
                                data = json.loads(json_str)
                                if not isinstance(data, list):
                                    return json_response(error="Invalid JSON format.")

                                patches_data.extend(data)
                
                # 将数据保存到数据库
                import_patch(patches_data)
                
                return json_response(data="Data imported successfully.")
            else:
                return json_response(error="Invalid request or file not provided.")
        except Exception as e:
            return json_response(error=str(e))


@auth('cve.host.scan')
def batch_host_scan(request): 
    if request.method != 'POST':
        return json_response(error='Invalid request method')

    data = json.loads(request.body)
    host_ids = data.get('host_ids')

    if not host_ids:
        hosts = Host.objects.all()
    else: 
        hosts = Host.objects.filter(id__in=host_ids)

    token = uuid.uuid4().hex
    Thread(target=do_batch_host_scan, args=(token, hosts)).start()
    return json_response({'token': token, 'hosts': {x.id: {'name': x.name} for x in hosts}})

@auth('cve.host.patch')
def batch_host_patch(request): 
    if request.method != 'POST':
        return json_response(error='Invalid request method')

    data = json.loads(request.body)
    host_ids = data.get('host_ids')
    cves = data.get('cves')
    
    if not host_ids:
        return json_response(error='host_ids is required')
    
    hosts = []
    for host_id in host_ids:
        host = Host.objects.filter(id=host_id).first()
        if host is None:
            return json_response(error='host not found')
        hosts.append(host)

    token = uuid.uuid4().hex
    Thread(target=do_batch_host_patch, args=(token, hosts, cves)).start()
    return json_response({'token': token, 'hosts': {x.id: {'name': x.name} for x in hosts}})


def get_cve_detail(request, cve):
    cve_detail = Cve.objects.filter(cve=cve).first()
    cve_detail = cve_detail.to_dict()

    cve_detail['related_hosts'] = []
    host_scan_cves = (HostScanCve.objects
                      .filter(cve=cve, is_fixed=False, hs__is_fixed=False)
                      .values('host_id', 'created_at'))
    host_scan_cves_dict = {x['host_id']: x['created_at'] for x in host_scan_cves}

    host_ids_set = set()
    host_ids_set.update(host_scan_cves_dict.keys())
    if not request.user.is_supper:
        host_ids_set = host_ids_set.intersection(get_host_perms(request.user))

    hosts = (Host.objects
             .filter(id__in=host_ids_set)
             .select_related('hostextend')
             .prefetch_related('groups')
             )
    for host in hosts:
        host_dict = host.to_dict()
        host_dict['group_names'] = ', '.join([g.name for g in host.groups.all()])
        host_dict['last_scan_at'] = host_scan_cves_dict.get(host.id)
        cve_detail['related_hosts'].append(host_dict)

    patches = Patch.objects.filter(cve=cve)
    package_name_set = set()
    for patch in patches:
        package_name_set.add(patch.package_name)
    cve_detail['packages'] = list(package_name_set)

    security_level = get_security_level(cve_detail['sa_cvss'])
    cve_detail['security_level'] = security_level
    return json_response(cve_detail)

def get_host_detail(request, host_id):
    host = Host.objects.filter(id=host_id).select_related('hostextend').prefetch_related("groups").first()
    host_dict = host.to_dict()
    host_dict['group_names'] = ', '.join([g.name for g in host.groups.all()])
    host_dict['os_name'] = host.hostextend.os_name

    host_scan = HostScan.objects.filter(host_id=host_id, is_fixed=False).first()
    if host_scan:
        host_dict['last_scan_at'] = host_scan.created_at

    all_cves = []
    host_dict['host_scan_cves'] = []
    host_scan_cves = HostScanCve.objects.filter(hs_id=host_scan.id, is_fixed=False)
    for host_scan_cve in host_scan_cves:
        host_scan_cve_dict = host_scan_cve.to_dict()
        host_scan_cve_dict['security_level'] = get_security_level(host_scan_cve.cvss)
        all_cves.append(host_scan_cve.cve)
        host_dict['host_scan_cves'].append(host_scan_cve_dict)
        security_level = get_security_level(host_scan_cve.cvss)
        host_dict[f'security_level_{security_level}_count'] = host_dict.get(f'security_level_{security_level}_count', 0) + 1

    host_dict['host_scan_patches'] = []
    host_scan_patches = HostScanPatch.objects.filter(cve__in=all_cves, is_fixed=False)
    for host_scan_patch in host_scan_patches:
        host_scan_patch_dict = host_scan_patch.to_dict()
        host_scan_patch_dict['security_level'] = get_security_level(host_scan_patch.cvss)
        host_dict['host_scan_patches'].append(host_scan_patch_dict)

    return json_response(host_dict)
